#include <iostream>
#include <iomanip>
#include <cassert>

#include "arguments.hxx"
#include "timer_cuda.cuh"
#include "cuda_check.cuh"

__global__ void initCuda(double* u, double value, size_t n)
{
  size_t i = blockDim.x * blockIdx.x + threadIdx.x;
  if (i < n)
    u[i] = (i>n/4 && i<3*n/4) ? value : 0.0;
}

void init(double *c, size_t n, double a, TimerCuda &T)
{
  double x = a/(n-1);

  T.markStart();
  initCuda<<<(n+512)/512, 512>>>(c, x, n);
  T.markStop();
}

__global__ void calculCuda1(double* u, double *v, size_t n)
{
  size_t i = blockDim.x * blockIdx.x + threadIdx.x;
  if (i > 0 && i < n-1)
    v[i] = (u[i-1] + 6*u[i] + u[i+1])/8.0;
}

void calcul1(double *u, double *v, size_t n, int p, TimerCuda &T)
{
  T.markStart();
  for (int i=0; i<p; i++) {
    calculCuda1<<<(n+512)/512, 512>>>(u, v, n);
    std::swap(u, v);
  }
  T.markStop();
}

__global__ void calculCuda2(double* u, double *v, size_t n)
{
  size_t i = blockDim.x * blockIdx.x + threadIdx.x;
  if (i > 0 && i < n-1)
    v[i] = (u[i-1] + 2*u[i] + u[i+1])/4.0;
}

void calcul2(double *u, double *v, size_t n, int p, TimerCuda &T)
{
  T.markStart();
  for (int i=0; i<p; i++) {
    calculCuda2<<<(n+512)/512, 512>>>(u, v, n);
    std::swap(u, v);
  }
  T.markStop();
}


int main(int argc, char** argv)
{
  Arguments Args(argc, argv);
  size_t n = Args.Get("n", 10000L);
  size_t m = Args.Get("m", n*2);

  std::cout << "\n\nn = " << n << "\n\n";

    TimerCuda T_ref;
    std::vector<TimerCuda> T(12);

    T_ref.markStart();

    double *u1, *u2, *v1, *v2;
    int k = 0;

    T[k].init(T_ref, "CudaMalloc_u1");
    T[k].markStart();
    checkCudaOp(cudaMalloc(&u1, n * sizeof(double)));
    T[k].markStop();

    k++;
    T[k].init(T_ref, "CudaMalloc_v1");
    T[k].markStart();
    checkCudaOp(cudaMalloc(&v1, n * sizeof(double)));
    T[k].markStop();

    k++;
    T[k].init(T_ref, "CudaMalloc_u2");
    T[k].markStart();
    checkCudaOp(cudaMalloc(&u2, m * sizeof(double)));
    T[k].markStop();

    k++;
    T[k].init(T_ref, "CudaMalloc_v2");
    T[k].markStart();
    checkCudaOp(cudaMalloc(&v2, m * sizeof(double)));
    T[k].markStop();

    k++;
    T[k].init(T_ref, "InitKernel1");
    init(u1, n, 1.0, T[k]);

    k++;
    T[k].init(T_ref, "InitKernel2");
    init(u2, m, 2.0, T[k]);

    k++;
    T[k].init(T_ref, "CalculKernel1");
    calcul1(u1, v1, n, 200, T[k]);

    k++;
    T[k].init(T_ref, "CalculKernel2");
    calcul2(u2, v2, m, 120, T[k]);

    k++;
    T[k].init(T_ref, "CudaFree_u1");
    T[k].markStart();
    checkCudaOp(cudaFree(u1));
    T[k].markStop();

    k++;
    T[k].init(T_ref, "CudaFree_v1");
    T[k].markStart();
    checkCudaOp(cudaFree(v1));
    T[k].markStop();

    k++;
    T[k].init(T_ref, "CudaFree_u2");
    T[k].markStart();
    checkCudaOp(cudaFree(u2));
    T[k].markStop();

    k++;
    T[k].init(T_ref, "CudaFree_v2");
    T[k].markStart();
    checkCudaOp(cudaFree(v2));
    T[k].markStop();

    k++;
    cudaDeviceSynchronize();

    for (int i=0; i<k; i++) {
      std::cout << T[i] << std::endl;
    }
  return 0;
}
