#include <iostream>
#include <iomanip>
#include <cassert>

#include "arguments.hxx"
#include "timer_cuda_stream.cuh"

__global__ void initCuda(double* u, double value, size_t n)
{
  size_t i = blockDim.x * blockIdx.x + threadIdx.x;
  if (i < n)
    u[i] = (i>n/4 && i<3*n/4) ? value : 0.0;
}

void init(double *c, size_t n, double a, cudaStream_t stream)
{
  double x = a/(n-1);

  initCuda<<<(n+512)/512, 512, 0, stream>>>(c, x, n);
}

__global__ void calculCuda1(double* u, double *v, size_t n)
{
  size_t i = blockDim.x * blockIdx.x + threadIdx.x;
  if (i > 0 && i < n-1)
    v[i] = (u[i-1] + 6*u[i] + u[i+1])/8.0;
}

void calcul1(double *u, double *v, size_t n, int p, cudaStream_t stream)
{
  for (int i=0; i<p; i++) {
    calculCuda1<<<(n+512)/512, 512, 0, stream>>>(u, v, n);
    std::swap(u, v);
  }
}

__global__ void calculCuda2(double* u, double *v, size_t n)
{
  size_t i = blockDim.x * blockIdx.x + threadIdx.x;
  if (i > 0 && i < n-1)
    v[i] = (u[i-1] + 2*u[i] + u[i+1])/4.0;
}

void calcul2(double *u, double *v, size_t n, int p, cudaStream_t stream)
{
  for (int i=0; i<p; i++) {
    calculCuda2<<<(n+512)/512, 512, 0, stream>>>(u, v, n);
    std::swap(u, v);
  }
}


int main(int argc, char** argv)
{
  Arguments Args(argc, argv);
  size_t n = Args.Get("n", 10000L);
  size_t m = Args.Get("m", n*2);

  std::cout << "\n\nn = " << n << "\n\n";
  
  cudaStream_t stream[2];
  cudaStreamCreate(&stream[0]);
  cudaStreamCreate(&stream[1]);

  TimerCudaStream T_ref;
  std::vector<TimerCudaStream> T(12);

  T_ref.markStart();

  double *u1, *u2, *v1, *v2;
  int k = 0;

  T[k].init(T_ref, stream[0], "CudaMalloc_u1", 0);
  T[k].markStart();
  cudaMallocAsync(&u1, n * sizeof(double), stream[0]);
  T[k].markStop();

  k++;
  T[k].init(T_ref, stream[0], "CudaMalloc_v1", 0);
  T[k].markStart();
  cudaMallocAsync(&v1, n * sizeof(double), stream[0]);
  T[k].markStop();

  k++;
  T[k].init(T_ref, stream[0], "InitKernel1", 0);
  T[k].markStart();
  init(u1, n, 1.0, stream[0]);
  T[k].markStop();

  k++;
  T[k].init(T_ref, stream[0], "CalculKernel1", 0);
  T[k].markStart();
  calcul1(u1, v1, n, 200, stream[0]);
  T[k].markStop();

  k++;
  T[k].init(T_ref, stream[0], "CudaFree_u1", 0);
  T[k].markStart();
  cudaFreeAsync(u1, stream[0]);
  T[k].markStop();

  k++;
  T[k].init(T_ref, stream[0], "CudaFree_v1", 0);
  T[k].markStart();
  cudaFreeAsync(v1, stream[0]);
  T[k].markStop();

  k++;
  T[k].init(T_ref, stream[1], "CudaMalloc_u2", 1);
  T[k].markStart();
  cudaMallocAsync(&u2, m * sizeof(double), stream[1]);
  T[k].markStop();

  k++;
  T[k].init(T_ref, stream[1], "CudaMalloc_v2", 1);
  T[k].markStart();
  cudaMallocAsync(&v2, m * sizeof(double), stream[1]);
  T[k].markStop();

  k++;
  T[k].init(T_ref, stream[1], "InitKernel2",1);
  T[k].markStart();
  init(u2, m, 2.0, stream[1]);
  T[k].markStop();

  k++;
  T[k].init(T_ref, stream[1], "CalculKernel2", 1);
  T[k].markStart();
  calcul2(u2, v2, m, 120, stream[1]);
  T[k].markStop();

  k++;
  T[k].init(T_ref, stream[1], "CudaFree_u2", 1);
  T[k].markStart();
  cudaFreeAsync(u2, stream[1]);
  T[k].markStop();

  k++;
  T[k].init(T_ref, stream[1], "CudaFree_v2", 1);
  T[k].markStart();
  cudaFreeAsync(v2, stream[1]);
  T[k].markStop();

  k++;

  cudaDeviceSynchronize();

  cudaStreamDestroy(stream[0]);
  cudaStreamDestroy(stream[1]);

  for (int i=0; i<k; i++) {
    std::cout << T[i] << std::endl;
  }

  return 0;
}
