#include <iostream>
#include <iomanip>
#include <cassert>

#include "Matrix.hxx"
#include "Vector.hxx"
#include "dot.cuh"
#include "cuda_check.cuh"
#include "arguments.hxx"
#include "timer.hxx"
#include "timer_cuda.cuh"
#include "dData.hxx"

const char *version = VERSION;

__global__ void initCuda(double* a, double value, size_t n)
{
  size_t i = blockDim.x * blockIdx.x + threadIdx.x;
  if (i < n) {
    a[i] = value * i;
  }
}

TimerCuda T[10];

void init(dVector &v, double a, TimerCuda &T)
{
  size_t n = v.n();
  double x = a/(n-1);
  double *c = v.data();

  T.markStart();
  initCuda<<<(n+512)/512, 512>>>(c, x, n);
  T.markStop();
  checkCudaKernel();
}

void init(dMatrix &M, double a, TimerCuda &T)
{
  size_t n = M.n(), m = M.m(), nm = n*m;
  double x = a/(nm-1);
  double *c = M.data();
  T.markStart();
  initCuda<<<(nm+512)/512, 512>>>(c, x, nm);
  T.markStop();
  checkCudaKernel();
}

__global__ void matvecCuda(double *w, const double *A, const double *v, size_t n, size_t m)
{
  size_t i = blockDim.x * blockIdx.x + threadIdx.x, j;
  double s = 0.0;
  if (i < n) {
    for (j=0; j<m; j++)
      s += A[i*m + j] * v[j];
    w[i] = s;
  }
}

void matvec(dVector &w, const dMatrix &A, const dVector &v, TimerCuda &T)
{
  size_t n = A.n(), m = A.m() ;
  assert(w.n() == A.n());
  assert(v.n() == A.m());
  const double *x = v.data();
  double * y = w.data();
  const double *M = A.data();

  T.markStart();
  matvecCuda<<<(n + 64)/64, 64>>>(y, M, x, n, m);
  T.markStop();
  checkCudaKernel();
}

double dot(const dVector & d_x,
           const dVector & d_y, TimerCuda &T)
{
  assert(d_y.n() == d_x.n());
  size_t n = d_x.n();
  size_t blockSize = 32;
  size_t gridSize = (n+blockSize)/blockSize;

  const double * dx = d_x.data(), * dy = d_y.data();
  double result =  dotCuda(n, dx, dy,
 	                   blockSize, gridSize, T);

  return result;
}

int main(int argc, char** argv)
{
  Arguments Args(argc, argv);
  size_t n = Args.Get("n", 10000L);
  size_t m = Args.Get("m", n);

  std::cout << "version " << version
            << "\n\nn,m = " << n << " " << m << "\n\n";

  Timer T_global;
  T_global.start();

  dData D(n, m);

  double s1, s2, s;

  T[0].name("init y");
  init(D.y, 1.5, T[0]);

  T[1].name("init A1");
  init(D.A1, 2.0, T[1]);

  T[2].name("init x1");
  init(D.x1, 1.0, T[2]);

  T[5].name("prod A1 x1");
  matvec(D.temp1, D.A1, D.x1, T[5]);
  T[6].name("dot (1ere partie)");
  s1 = dot(D.y, D.temp1, T[6]);

  T[3].name("init x1");
  init(D.A2, 3.0,T[3]);
  T[4].name("init x1");
  init(D.x2, 1.2,T[4]);

  T[7].name("prod A2 x2");
  matvec(D.temp2, D.A2, D.x2, T[7]);
  T[8].name("dot (2ème partie)");
  s2 = dot(D.y, D.temp2, T[8]);

  s = s1 + s2;

  std::cout << "\n\ty (A1 x1 + A2 x2) = " << s << std::endl;

  if (n <= 10 && m <= 10) {
    D.x1.print();
    D.A1.print();
    D.x2.print();
    D.A2.print();
    D.y.print();
  }
  std::cout << "\n";
  T_global.stop();

  for (auto iT:T)
    std::cout << iT << std::endl;

  std::cout << "T global (" << version << ") : " << T_global.elapsed() << " s" << std::endl;
  return 0;
}
