#include <iostream>
#include <iomanip>
#include <cassert>

#include "Matrix.hxx"
#include "Vector.hxx"
#include "dot.h"
#include "arguments.hxx"
#include "timer.hxx"
#include "dData.hxx"

#include "timer_cuda.cuh"
#include "cuda_check.cuh"
#include "stream_check.cuh"

const char *version = VERSION;

__global__ void initCuda(double* a, double value, size_t n)
{
  size_t i = blockDim.x * blockIdx.x + threadIdx.x;
  if (i < n) {
    a[i] = value * i;
  }
}

void init(dVector &v, double a)
{
  size_t n = v.n();
  double x = a/(n-1);
  double *c = v.data();

  initCuda<<<(n+512)/512, 512>>>(c, x, n);
  checkCudaKernel();
}

void init(dVector &v, double a,
          cudaStream_t stream)
{
  size_t n = v.n();
  double x = a/(n-1);
  double *c = v.data();

  initCuda<<<(n+512)/512, 512, 0, stream>>>(c, x, n);
  checkStreamKernel(stream);
}

void init(dMatrix &M, double a,
          cudaStream_t stream)
{
  size_t n = M.n(), m = M.m(), nm = n*m;
  double x = a/(nm-1);
  double *c = M.data();
  initCuda<<<(nm+512)/512, 512, 0, stream>>>(c, x, nm);
  checkStreamKernel(stream);
}

__global__ void matvecCuda(double *w, const double *A, const double *v, size_t n, size_t m)
{
  size_t i = blockDim.x * blockIdx.x + threadIdx.x, j;
  double s = 0.0;
  if (i < n) {
    for (j=0; j<m; j++)
      s += A[i*m + j] * v[j];
    w[i] = s;
  }
}

void matvec(dVector &w, const dMatrix &A, const dVector &v,
            cudaStream_t stream)
{
  size_t n = A.n(), m = A.m() ;
  assert(w.n() == A.n());
  assert(v.n() == A.m());
  const double *x = v.data();
  double * y = w.data();
  const double *M = A.data();

  matvecCuda<<<(n + 64)/64, 64, 0, stream>>>(y, M, x, n, m);
  checkStreamKernel(stream);
}

double dot(const dVector & d_x,
           const dVector & d_y,
           cudaStream_t stream)
{
  assert(d_y.n() == d_x.n());
  size_t n = d_x.n();
  size_t blockSize = 32;
  size_t gridSize = (n+blockSize)/blockSize;

  double result =  dotCuda(n, d_x.data(), d_y.data(),
 	                   blockSize, gridSize, stream);

  return result;
}

int main(int argc, char** argv)
{
  Arguments Args(argc, argv);
  size_t n = Args.Get("n", 10000L);
  size_t m = Args.Get("m", n);

  std::cout << "version " << version
            << "\n\nn,m = " << n << " " << m << "\n\n";

  Timer T_global;
  T_global.start();

  dData D(n, m);

  double s1, s2, s;

  cudaStream_t stream[2];
  checkCudaOp(cudaStreamCreate(&stream[0]));
  checkCudaOp(cudaStreamCreate(&stream[1]));

  Timer T_calcul;
  T_calcul.start();

  init(D.y, 1.5);
  cudaDeviceSynchronize();

#pragma omp sections 
  { 
   #pragma omp section
    {
      init(D.A1, 2.0, stream[0]);
      init(D.x1, 1.0, stream[0]);
      matvec(D.temp1, D.A1, D.x1, stream[0]);
      s1 = dot(D.y, D.temp1, stream[0]);
    }
   #pragma omp section 
    {
      init(D.A2, 3.0, stream[1]);
      init(D.x2, 1.2, stream[1]);
      matvec(D.temp2, D.A2, D.x2, stream[1]);
      s2 = dot(D.y, D.temp2, stream[1]);
    }
  }
  
  cudaDeviceSynchronize();
  s = s1 + s2;
  T_calcul.stop();

  std::cout << "\n\ty (A1 x1 + A2 x2) = " << s << std::endl;

  if (n <= 10 && m <= 10) {
    D.x1.print();
    D.A1.print();
    D.x2.print();
    D.A2.print();
    D.y.print();
  }

  checkCudaOp(cudaStreamDestroy(stream[0]));
  checkCudaOp(cudaStreamDestroy(stream[1]));
  
  std::cout << "\n";
  T_global.stop();
  std::cout << "T calcul (" << version << ") : " << T_calcul.elapsed() << " s" << std::endl;
  std::cout << "T global (" << version << ") : " << T_global.elapsed() << " s" << std::endl;
  return 0;
}
